import rdflib
from rdflib.namespace import (
    RDFS,
    RDF,
    OWL,
)
from rdflib import BNode, URIRef, Literal
from model.kgcl import (
    NodeRename,
    NodeMove,
    NodeDeepening,
    NodeShallowing,
    EdgeCreation,
    EdgeDeletion,
    PredicateChange,
    NodeCreation,
    ClassCreation,
    PlaceUnder,
    RemoveUnder,
    NewSynonym,
    RemovedNodeFromSubset,
)

from diff.change_detection import (
    detect_renamings,
    detect_node_moves,
    detect_predicate_changes,
)

from diff.render_operations import render

from diff.graph_diff import (
    get_added_thin_triples,
    get_deleted_thin_triples,
)


def id_generator():
    id = 0
    while True:
        yield id
        id += 1


id_gen = id_generator()


class SingleTripleChangeSummary:
    """
    Dataclass holding information about (atomic) existential restriction changes.
    """

    def __init__(self):

        # KGCL data
        self.renamings = []
        self.class_creations = []
        self.subsumption_creations = []
        self.subsumption_deletions = []
        self.predicate_changes = []
        self.node_moves = []
        self.synonym_creations = []
        self.edge_creations = []
        self.edge_deletions = []

        # RDF data
        self.covered_triples_renamings = []
        self.covered_triples_class_creations = []
        self.covered_triples_subsumption_creations = []
        self.covered_triples_subsumption_deletions = []
        self.covered_triples_predicate_changes = []
        self.covered_triples_node_moves = []
        self.covered_triples_synonym_creations = []
        self.covered_triples_edge_creations = []
        self.covered_triples_edge_deletions = []

        # non-deterministic data
        self.non_deterministic_node_moves = []
        self.non_deterministic_predicate_changes = []
        self.non_deterministic_renamings = []

    def get_commands(self):
        kgcl_commands = []
        for k in self.synonym_creations:
            kgcl_commands.append(k)
        for k in self.node_moves:
            kgcl_commands.append(k)
        for k in self.predicate_changes:
            kgcl_commands.append(k)
        for k in self.edge_creations:
            kgcl_commands.append(k)
        for k in self.edge_deletions:
            kgcl_commands.append(k)
        for k in self.renamings:
            kgcl_commands.append(k)
        for k in self.class_creations:
            kgcl_commands.append(k)
        for k in self.subsumption_creations:
            kgcl_commands.append(k)
        for k in self.subsumption_deletions:
            kgcl_commands.append(k)

        return kgcl_commands

    def get_renamings(self):
        return self.renamings

    def get_class_creations(self):
        return self.class_creations

    def get_subsumption_creations(self):
        return self.subsumption_creations

    def get_subsumption_deletions(self):
        return self.subsumption_deletions

    def get_predicate_changes(self):
        return self.predicate_changes

    def get_node_moves(self):
        return self.node_moves

    def get_synonym_creations(self):
        return self.synonym_creations

    def get_edge_creations(self):
        return self.edge_creations

    def get_edge_deletions(self):
        return self.edge_deletions

    def get_summary_KGCL_commands(self):
        out = (
            "Renamings: "
            + str(len(self.renamings))
            + "\n"
            + "ClassCreations: "
            + str(len(self.class_creations))
            + "\n"
            # + "Subsumption Creations: "
            + "PlaceUnder: "
            + str(len(self.subsumption_creations))
            + "\n"
            # + "Subsumption Deletions: "
            + "RemoveUnder: "
            + str(len(self.subsumption_deletions))
            + "\n"
            + "PredicateChanges: "
            + str(len(self.predicate_changes))
            + "\n"
            + "NodeMoves: "
            + str(len(self.node_moves))
            + "\n"
            + "SynonymCreations: "
            + str(len(self.synonym_creations))
            + "\n"
            + "EdgeCreations: "
            + str(len(self.edge_creations))
            + "\n"
            + "EdgeDeletions: "
            + str(len(self.edge_deletions))
            + "\n"
        )
        return out

    def get_summary_RDF_triples(self):
        out = (
            "Renamings:"
            + str(len(self.covered_triples_renamings))
            + "\n"
            + "ClassCreations: "
            + str(len(self.covered_triples_class_creations))
            + "\n"
            # + "Subsumption Creations: "
            + "PlaceUnder: "
            + str(len(self.covered_triples_subsumption_creations))
            + "\n"
            # + "Subsumption Deletions: "
            + "RemoveUnder: "
            + str(len(self.covered_triples_subsumption_deletions))
            + "\n"
            + "PredicateChanges: "
            + str(len(self.covered_triples_predicate_changes))
            + "\n"
            + "NodeMoves: "
            + str(len(self.covered_triples_node_moves))
            + "\n"
            + "SynonymCreations: "
            + str(len(self.covered_triples_synonym_creations))
            + "\n"
            + "EdgeCreations: "
            + str(len(self.covered_triples_edge_creations))
            + "\n"
            + "EdgeDeletions: "
            + str(len(self.covered_triples_edge_deletions))
            + "\n"
        )
        return out

    # RDF data

    def get_covered_triples_renamings(self):
        return self.covered_triples_renamings

    def add_covered_triples_renamings(self, triples):
        for t in triples:
            self.covered_triples_renamings.append(t)

    def get_covered_triples_class_creations(self):
        return self.covered_triples_class_creations

    def add_covered_triples_class_creations(self, triples):
        for t in triples:
            self.covered_triples_class_creations.append(t)

    def get_covered_triples_subsumption_creations(self):
        return self.covered_triples_subsumption_creations

    def add_covered_triples_subsumption_creations(self, triples):
        for t in triples:
            self.covered_triples_subsumption_creations.append(t)

    def get_covered_triples_subsumption_deletions(self):
        return self.covered_triples_subsumption_deletions

    def add_covered_triples_subsumption_deletions(self, triples):
        for t in triples:
            self.covered_triples_subsumption_deletions.append(t)

    def get_covered_triples_predicate_changes(self):
        return self.covered_triples_predicate_changes

    def add_covered_triples_predicate_changes(self, triples):
        for t in triples:
            self.covered_triples_predicate_changes.append(t)

    def get_covered_triples_node_moves(self):
        return self.covered_triples_node_moves

    def add_covered_triples_node_moves(self, triples):
        for t in triples:
            self.covered_triples_node_moves.append(t)

    def get_covered_triples_synonym_creations(self):
        return self.covered_triples_synonym_creations

    def add_covered_triples_synonym_creations(self, triples):
        for t in triples:
            self.covered_triples_synonym_creations.append(t)

    def get_covered_triples_edge_creations(self):
        return self.covered_triples_edge_creations

    def add_covered_triples_edge_creations(self, triples):
        for t in triples:
            self.covered_triples_edge_creations.append(t)

    def get_covered_triples_edge_deletions(self):
        return self.covered_triples_edge_deletions

    def add_covered_triples_edge_deletions(self, triples):
        for t in triples:
            self.covered_triples_edge_deletions.append(t)

    # KGCL data

    def add_edge_creations(self, ls):
        for i in ls:
            self.edge_creations.append(i)

    def add_edge_deletions(self, ls):
        for i in ls:
            self.edge_deletions.append(i)

    def add_renamings(self, ls):
        for i in ls:
            self.renamings.append(i)

    def add_class_creations(self, ls):
        for i in ls:
            self.class_creations.append(i)

    def add_subsumption_creations(self, ls):
        for i in ls:
            self.subsumption_creations.append(i)

    def add_subsumption_deletions(self, ls):
        for i in ls:
            self.subsumption_deletions.append(i)

    def add_edge_creation(self, i):
        self.edge_creations.append(i)

    def add_edge_deletion(self, i):
        self.edge_deletions.append(i)

    def add_renaming(self, i):
        self.renamings.append(i)

    def add_class_creation(self, i):
        self.class_creations.append(i)

    def add_subsumption_creation(self, i):
        self.subsumption_creations.append(i)

    def add_subsumption_deletion(self, i):
        self.subsumption_deletions.append(i)

    def add_predicate_change(self, i):
        self.predicate_changes.append(i)

    def add_node_move(self, i):
        self.node_moves.append(i)

    def add_synonym_creation(self, i):
        self.synonym_creations.append(i)

    # non-deterministic data
    def add_non_deterministic_node_moves(self, ls):
        self.non_deterministic_node_moves += ls

    def add_non_deterministic_predicate_changes(self, ls):
        self.non_deterministic_predicate_changes += ls

    def add_non_deterministic_renamings(self, ls):
        self.non_deterministic_renamings += ls

    def get_non_deterministic_node_moves(self):
        return self.non_deterministic_node_moves

    def get_non_deterministic_predicate_changes(self):
        return self.non_deterministic_predicate_changes

    def get_non_deterministic_renamings(self):
        return self.non_deterministic_renamings


def generate_thin_triple_commands(g1, g2):
    """
    Given two graphs g1 and g2,
    return all KGCL data model instances to account for their diff.
    """
    # summary object for single triple changes
    summary = SingleTripleChangeSummary()

    added = get_added_thin_triples(g1, g2)
    deleted = get_deleted_thin_triples(g1, g2)

    # synonyms
    synonym_additions, covered = generate_synonym_creations(added)
    summary.add_covered_triples_synonym_creations(covered)
    added = added - covered
    # TODO: extend data model for deleted synonyms
    # synonym_deletions, covered = generate_synonym_deletions(added)

    # deepen [TODO: need some kind of reasoning/querying for this]
    # shallow [TODO: need some kind of reasoning/querying for this]

    # move
    node_moves, covered, non_deterministic = detect_node_moves(added, deleted)
    summary.add_covered_triples_node_moves(covered)
    summary.add_non_deterministic_node_moves(non_deterministic)
    added = added - covered
    deleted = deleted - covered

    # change relationship
    relationship_change, covered, non_deterministic = detect_predicate_changes(
        added, deleted
    )
    summary.add_covered_triples_predicate_changes(covered)
    summary.add_non_deterministic_predicate_changes(non_deterministic)
    added = added - covered
    deleted = deleted - covered

    # renamings
    renamings, covered, non_deterministic = detect_renamings(added, deleted)
    summary.add_covered_triples_renamings(covered)
    summary.add_non_deterministic_renamings(non_deterministic)
    added = added - covered
    deleted = deleted - covered

    # subsumptions
    subsumption_creations, covered = generate_subsumption_creations(added)
    summary.add_covered_triples_subsumption_creations(covered)
    added = added - covered

    subsumption_deletions, covered = generate_subsumption_deletions(deleted)
    summary.add_covered_triples_subsumption_deletions(covered)
    deleted = deleted - covered

    # create node
    node_creations, covered = generate_class_creations(added)
    summary.add_covered_triples_class_creations(covered)
    added = added - covered

    # everything else
    edge_creations, covered = generate_edge_creations(added)
    summary.add_covered_triples_edge_creations(covered)
    edge_deletions, covered = generate_edge_deletions(deleted)
    summary.add_covered_triples_edge_deletions(covered)

    for s in synonym_additions:
        summary.add_synonym_creation(render(s))
    for c in node_creations:
        summary.add_class_creation(render(c))
    for m in node_moves:
        summary.add_node_move(render(m))
    for r in relationship_change:
        summary.add_predicate_change(render(r))
    for s in subsumption_creations:
        summary.add_subsumption_creation(render(s))
    for s in subsumption_deletions:
        summary.add_subsumption_deletion(render(s))
    for r in renamings:
        summary.add_renaming(render(r))
    for e in edge_creations:
        summary.add_edge_creation(render(e))
    for e in edge_deletions:
        summary.add_edge_deletion(render(e))

    return summary


def get_type(rdf_entity):
    if isinstance(rdf_entity, URIRef):
        return "uri"
    elif isinstance(rdf_entity, Literal):
        return "literal"
    else:
        return "Error"


def get_language_tag(rdf_entity):
    if isinstance(rdf_entity, Literal):
        return rdf_entity.language
    else:
        return None


def get_datatype(rdf_entity):
    if isinstance(rdf_entity, Literal) and rdf_entity.datatype is not None:
        return str(rdf_entity.datatype)
    else:
        return None


# TODO: identify node creations + labels
def generate_class_creations(added):
    covered = rdflib.Graph()
    kgcl = []

    for s, p, o in added.triples((None, RDF.type, OWL.Class)):
        id = "test_id_" + str(next(id_gen))
        covered.add((s, p, o))
        node = ClassCreation(id=id, about_node=str(s), node_id=str(s))
        kgcl.append(node)

    return kgcl, covered


def generate_synonym_creations(added):
    covered = rdflib.Graph()
    kgcl = []

    synonym = URIRef("http://www.geneontology.org/formats/oboInOwl#hasSynonym")
    exact_synonym = URIRef(
        "http://www.geneontology.org/formats/oboInOwl#hasExactSynonym"
    )
    narrow_synonym = URIRef(
        "http://www.geneontology.org/formats/oboInOwl#hasNarrowSynonym"
    )
    broad_synonym = URIRef(
        "http://www.geneontology.org/formats/oboInOwl#hasBroadSynonym"
    )
    related_synonym = URIRef(
        "http://www.geneontology.org/formats/oboInOwl#hasRelatedSynonym"
    )

    include = {synonym, exact_synonym, narrow_synonym, broad_synonym, related_synonym}

    # synonyms
    for s, p, o in added:
        if p in include:
            id = "test_id_" + str(next(id_gen))
            covered.add((s, p, o))

            language = get_language_tag(o)
            # datatype = get_datatype(o)

            qualifier = ""  # case for hasSynonym
            if p == exact_synonym:
                qualifier = "exact"
            elif p == narrow_synonym:
                qualifier = "narrow"
            elif p == broad_synonym:
                qualifier = "broad"
            elif p == related_synonym:
                qualifier = "related"

            node = NewSynonym(
                id=id,
                about_node=str(s),
                new_value=str(o),
                qualifier=qualifier,
                language=language,
            )

            kgcl.append(node)

    return kgcl, covered


# TODO: extend data model for deleted synonyms
def generate_synonym_deletions(deleted):
    covered = rdflib.Graph()
    kgcl = []

    # synonym = URIRef("http://www.geneontology.org/formats/oboInOwl#hasSynonym")
    exact_synonym = URIRef(
        "http://www.geneontology.org/formats/oboInOwl#hasExactSynonym"
    )
    narrow_synonym = URIRef(
        "http://www.geneontology.org/formats/oboInOwl#hasNarrowSynonym"
    )
    broad_synonym = URIRef(
        "http://www.geneontology.org/formats/oboInOwl#hasBroadSynonym"
    )
    related_synonym = URIRef(
        "http://www.geneontology.org/formats/oboInOwl#hasRelatedSynonym"
    )

    # synonyms
    for s, p, o in deleted:
        id = "test_id_" + str(next(id_gen))
        covered.add((s, p, o))

        language = get_language_tag(o)
        # datatype = get_datatype(o)

        qualifier = ""  # case for hasSynonym
        if p == exact_synonym:
            qualifier = "exact"
        elif p == narrow_synonym:
            qualifier = "narrow"
        elif p == broad_synonym:
            qualifier = "broad"
        elif p == related_synonym:
            qualifier = "related"

        # TODO: extend data model for deleted synonyms
        node = NewSynonym(
            id=id,
            about_node=str(s),
            new_value=str(o),
            qualifier=qualifier,
            language=language,
        )

        kgcl.append(node)

    return kgcl, covered


def generate_subsumption_creations(added):
    covered = rdflib.Graph()
    kgcl = []

    for s, p, o in added.triples((None, RDFS.subClassOf, None)):
        id = "test_id_" + str(next(id_gen))
        subclass = str(s)
        superclass = str(o)

        # TODO the hardcoded owl:subClassOf should be part of the data model
        node = PlaceUnder(
            id=id,
            subject=subclass,
            predicate="<http://www.w3.org/2000/01/rdf-schema#subClassOf>",
            object=superclass,
        )

        kgcl.append(node)
        covered.add((s, p, o))

    return kgcl, covered


def generate_subsumption_deletions(deleted):
    covered = rdflib.Graph()
    kgcl = []

    for s, p, o in deleted.triples((None, RDFS.subClassOf, None)):
        id = "test_id_" + str(next(id_gen))
        subclass = str(s)
        superclass = str(o)

        # TODO the hardcoded owl:subClassOf should be part of the data model
        node = RemoveUnder(
            id=id,
            subject=subclass,
            predicate="<http://www.w3.org/2000/01/rdf-schema#subClassOf>",
            object=superclass,
        )

        kgcl.append(node)
        covered.add((s, p, o))

    return kgcl, covered


def generate_edge_creations(added):
    covered = rdflib.Graph()

    kgcl = []
    for s, p, o in added:

        id = "test_id_" + str(next(id_gen))
        object_type = get_type(o)
        language_tag = get_language_tag(o)
        datatype = get_datatype(o)
        if datatype is not None:
            datatype = "<" + datatype + ">"  # expect data types without curies

        node = EdgeCreation(
            id=id,
            subject=str(s),
            predicate=str(p),
            object=str(o),
            object_type=str(object_type),
            language=language_tag,
            datatype=datatype,
        )

        kgcl.append(node)
        covered.add((s, p, o))

    return kgcl, covered


def generate_edge_deletions(deleted):
    covered = rdflib.Graph()
    kgcl = []
    for s, p, o in deleted:

        id = "test_id_" + str(next(id_gen))
        object_type = get_type(o)
        language_tag = get_language_tag(o)
        datatype = get_datatype(o)
        if datatype is not None:
            datatype = "<" + datatype + ">"

        node = EdgeDeletion(
            id=id,
            subject=str(s),
            predicate=str(p),
            object=str(o),
            object_type=object_type,
            language=language_tag,
            datatype=datatype,
        )

        kgcl.append(node)
        covered.add((s, p, o))

    return kgcl, covered



from lark import Lark, Token
import re
from model.kgcl import (
    NodeRename,
    NodeObsoletion,
    NodeUnobsoletion,
    NodeDeletion,
    NodeMove,
    NodeDeepening,
    NodeShallowing,
    EdgeCreation,
    EdgeDeletion,
    PredicateChange,
    NodeCreation,
    ClassCreation,
    NewSynonym,
    RemovedNodeFromSubset,
    PlaceUnder,
    RemoveUnder,
    ExistentialRestrictionCreation,
    ExistentialRestrictionDeletion,
)
from model.ontology_model import Edge, Annotation
from pathlib import Path


def id_generator():
    """
    Returns a new ID for KGCL change operations.
    """
    id = 0
    while True:
        yield id
        id += 1


# initialise ID generator
id_gen = id_generator()


# initialise parser
path = Path(__file__).parent
kgcl_parser = Lark.open(str(path) + "/kgcl.lark", start="expression")


def parse(input):
    """
    Parse a set of KGCL command separated by \n and
    return instantiated dataclass objects from model.kgcl.
    """
    statements = input.splitlines()
    parsed = []

    for s in statements:
        parsed.append(parse_statement(s))
    return parsed


def parse_statement(input):
    """
    Parse a KGCL command and
    return an instantiated dataclass object from model.kgcl.
    """

    tree = kgcl_parser.parse(input)
    id = "kgcl_change_id_" + str(next(id_gen))

    command = tree.data

    if command == "rename":
        return parse_rename(tree, id)
    elif command == "obsolete":
        return parse_obsolete(tree, id)
    elif command == "unobsolete":
        return parse_unobsolete(tree, id)
    elif command == "delete":
        return parse_delete(tree, id)
    elif command == "move":
        return parse_move(tree, id)
    elif command == "deepen":
        return parse_deepen(tree, id)
    elif command == "shallow":
        return parse_shallow(tree, id)
    elif command == "create_edge":
        return parse_create_edge(tree, id)
    elif command == "delete_edge":
        return parse_delete_edge(tree, id)
    elif command == "create_annotated_edge":
        return parse_create_annotated_edge(tree, id)
    elif command == "delete_annotated_edge":
        return parse_delete_annotated_edge(tree, id)
    elif command == "change_relationship":
        return parse_change_relationship(tree, id)
    elif command == "create":
        return parse_create(tree, id)
    elif command == "create_class":
        return parse_create_class(tree, id)
    elif command == "create_synonym":
        return parse_create_synonym(tree, id)
    elif command == "add_subsumption_axiom":
        return parse_add_subsumption_axiom(tree, id)
    elif command == "delete_subsumption_axiom":
        return parse_delete_subsumption_axiom(tree, id)
    elif command == "add_existential_restriction_axiom":
        return parse_add_existential_restriction_axiom(tree, id)
    elif command == "delete_existential_restriction_axiom":
        return parse_delete_existential_restriction_axiom(tree, id)
    elif command == "remove_from_subset":
        return parse_remove_from_subset(tree, id)
    else:
        raise NotImplementedError("No implementation for KGCL command: " + command)

    # TODO: does not have a field for subsets
    # if(command == "add_to_subset"):

    #    term_id = next(tree.find_data('id'))
    #    term_id_token = next(get_tokens(term_id))

    #    subset_id = next(tree.find_data('subset'))
    #    subset_id_token = next(get_tokens(subset_id))

    #    return AddNodeToSubset(
    #        id=id, in_subset=subset_id_token, about_node=term_id_token
    #     )

    # TODO: more commands
    # if(command == "merge"):
    # if(command == "add"):
    # if(command == "add_parent"):
    # if(command == "add_annotation"):
    # if(command == "add_axiom"): (in Manchester Syntax?)
    # if(command == "add_class"):
    # if(command == "add_text_definition"):
    # if(command == "merge"):


def parse_remove_from_subset(tree, id):
    term_id_token = extract(tree, "id")
    subset_id_token = extract(tree, "subset")

    return RemovedNodeFromSubset(
        id=id, subset=subset_id_token, about_node=term_id_token
    )


def parse_delete_existential_restriction_axiom(tree, id):
    subclass_token = extract(tree, "subclass")
    property_token = extract(tree, "property")
    filler_token = extract(tree, "filler")

    subclass, s_representation = get_entity_representation(subclass_token)
    property, p_representation = get_entity_representation(property_token)
    filler, f_representation = get_entity_representation(filler_token)

    return ExistentialRestrictionDeletion(
        id=id,
        subclass=subclass,
        property=property,
        filler=filler,
        subclass_type=s_representation,
        property_type=p_representation,
        filler_type=f_representation,
    )


def parse_add_existential_restriction_axiom(tree, id):
    subclass_token = extract(tree, "subclass")
    property_token = extract(tree, "property")
    filler_token = extract(tree, "filler")

    subclass, s_representation = get_entity_representation(subclass_token)
    property, p_representation = get_entity_representation(property_token)
    filler, f_representation = get_entity_representation(filler_token)

    return ExistentialRestrictionCreation(
        id=id,
        subclass=subclass,
        property=property,
        filler=filler,
        subclass_type=s_representation,
        property_type=p_representation,
        filler_type=f_representation,
    )


def parse_delete_subsumption_axiom(tree, id):
    subclass_token = extract(tree, "subclass")
    superclass_token = extract(tree, "superclass")

    subclass, sub_representation = get_entity_representation(subclass_token)
    superclass, sup_representation = get_entity_representation(superclass_token)

    # TODO the hardcoded owl:subClassOf should be part of the data model
    return RemoveUnder(
        id=id,
        subject=subclass,
        predicate="<http://www.w3.org/2000/01/rdf-schema#subClassOf>",
        object=superclass,
        subject_type=sub_representation,
        object_type=sup_representation,
    )


def parse_add_subsumption_axiom(tree, id):
    subclass_token = extract(tree, "subclass")
    superclass_token = extract(tree, "superclass")

    subclass, sub_representation = get_entity_representation(subclass_token)
    superclass, sup_representation = get_entity_representation(superclass_token)

    # TODO the hardcoded owl:subClassOf should be part of the data model
    return PlaceUnder(
        id=id,
        subject=subclass,
        predicate="<http://www.w3.org/2000/01/rdf-schema#subClassOf>",
        object=superclass,
        subject_type=sub_representation,
        object_type=sup_representation,
    )


def parse_create_synonym(tree, id):
    entity_token = extract(tree, "entity")
    entity, representation = get_entity_representation(entity_token)

    synonym_string_token = extract(tree, "synonym")
    synonym, representation2 = get_entity_representation(synonym_string_token)

    language_token = extract(tree, "language")
    qualifier_token = extract(tree, "synonym_qualifier")

    return NewSynonym(
        id=id,
        about_node=entity,
        about_node_representation=representation,
        new_value=synonym,
        qualifier=qualifier_token,
        language=language_token,
    )


def parse_create_class(tree, id):
    term_id_token = extract(tree, "id")
    entity, representation = get_entity_representation(term_id_token)

    return ClassCreation(
        id=id, node_id=entity, about_node_representation=representation
    )


# the KGCL model suggests the command
# 'create node {id} {label} with {annotation set}'
# TODO: handling of {annotation set}
def parse_create(tree, id):
    term_id_token = extract(tree, "id")
    label_token = extract(tree, "label")
    language_token = extract(tree, "language")

    entity, representation = get_entity_representation(term_id_token)

    return NodeCreation(
        id=id,
        about_node=entity,
        about_node_representation=representation,
        node_id=term_id_token,
        name=label_token,
        language=language_token,
    )


def parse_change_relationship(tree, id):
    subject_token = extract(tree, "entity_subject")
    object_token = extract(tree, "entity_object")

    subject, s_representation = get_entity_representation(subject_token)
    object, o_representation = get_entity_representation(object_token)

    old_token = extract(tree, "old_entity")
    new_token = extract(tree, "new_entity")

    old, old_representation = get_entity_representation(old_token)
    new, new_representation = get_entity_representation(new_token)

    edge = Edge(
        subject=subject,
        predicate=old,
        object=object,
        subject_representation=s_representation,
        predicate_representation=old_representation,
        object_representation=o_representation,
    )

    language_token = extract(tree, "language")
    datatype_token = extract(tree, "datatype")

    return PredicateChange(
        id=id,
        about_edge=edge,
        old_value=old,
        new_value=new,
        old_value_type=old_representation,
        new_value_type=new_representation,
        language=language_token,
        datatype=datatype_token,
    )


def parse_delete_annotated_edge(tree, id):
    subject_token = extract(tree, "entity_subject")
    predicate_token = extract(tree, "entity_predicate")
    object_token = extract(tree, "entity_object")

    subject, s_representation = get_entity_representation(subject_token)
    predicate, p_representation = get_entity_representation(predicate_token)
    object, o_representation = get_entity_representation(object_token)

    annotation_property_token = extract(tree, "annotation_property")
    annotation_token = extract(tree, "annotation")

    annotation_property, prop_representation = get_entity_representation(
        annotation_property_token
    )
    annotation, a_representation = get_entity_representation(annotation_token)

    annotation_set = Annotation(
        property=annotation_property,
        filler=annotation,
        property_type=prop_representation,
        filler_type=a_representation,
    )

    return EdgeDeletion(
        id=id,
        subject=subject,
        predicate=predicate,
        object=object,
        subject_type=s_representation,
        predicate_type=p_representation,
        object_type=o_representation,
        annotation_set=annotation_set,
    )


def parse_create_annotated_edge(tree, id):
    subject_token = extract(tree, "entity_subject")
    predicate_token = extract(tree, "entity_predicate")
    object_token = extract(tree, "entity_object")

    subject, s_representation = get_entity_representation(subject_token)
    predicate, p_representation = get_entity_representation(predicate_token)
    object, o_representation = get_entity_representation(object_token)

    annotation_property_token = extract(tree, "annotation_property")
    annotation_token = extract(tree, "annotation")

    annotation_property, prop_representation = get_entity_representation(
        annotation_property_token
    )
    annotation, a_representation = get_entity_representation(annotation_token)

    annotation_set = Annotation(
        property=annotation_property,
        filler=annotation,
        property_type=prop_representation,
        filler_type=a_representation,
    )

    return EdgeCreation(
        id=id,
        subject=subject,
        predicate=predicate,
        object=object,
        subject_type=s_representation,
        predicate_type=p_representation,
        object_type=o_representation,
        annotation_set=annotation_set,
    )


def parse_delete_edge(tree, id):
    subject_token = extract(tree, "entity_subject")
    predicate_token = extract(tree, "entity_predicate")
    object_token = extract(tree, "entity_object")

    subject, s_representation = get_entity_representation(subject_token)
    predicate, p_representation = get_entity_representation(predicate_token)
    object, o_representation = get_entity_representation(object_token)

    language_token = extract(tree, "language")
    datatype_token = extract(tree, "datatype")

    return EdgeDeletion(
        id=id,
        subject=subject,
        predicate=predicate,
        object=object,
        subject_type=s_representation,
        predicate_type=p_representation,
        object_type=o_representation,
        language=language_token,
        datatype=datatype_token,
    )


def parse_create_edge(tree, id):
    subject_token = extract(tree, "entity_subject")
    predicate_token = extract(tree, "entity_predicate")
    object_token = extract(tree, "entity_object")

    subject, s_representation = get_entity_representation(subject_token)
    predicate, p_representation = get_entity_representation(predicate_token)
    object, o_representation = get_entity_representation(object_token)

    language_token = extract(tree, "language")
    datatype_token = extract(tree, "datatype")

    return EdgeCreation(
        id=id,
        subject=subject,
        predicate=predicate,
        object=object,
        subject_type=s_representation,
        predicate_type=p_representation,
        object_type=o_representation,
        language=language_token,
        datatype=datatype_token,
    )


def parse_shallow(tree, id):
    entity_token = extract(tree, "entity")
    old_token = extract(tree, "old_entity")
    new_token = extract(tree, "new_entity")

    entity, e_representation = get_entity_representation(entity_token)
    old_entity, o_representation = get_entity_representation(old_token)
    new_entity, n_representation = get_entity_representation(new_token)

    edge = Edge(
        subject=entity,
        object=old_entity,
        subject_representation=e_representation,
        object_representation=o_representation,
    )

    return NodeShallowing(
        id=id,
        about_edge=edge,
        old_value=old_entity,
        new_value=new_entity,
        old_object_type=o_representation,
        new_object_type=n_representation,
    )


def parse_deepen(tree, id):
    entity_token = extract(tree, "entity")
    old_token = extract(tree, "old_entity")
    new_token = extract(tree, "new_entity")

    entity, e_representation = get_entity_representation(entity_token)
    old_entity, o_representation = get_entity_representation(old_token)
    new_entity, n_representation = get_entity_representation(new_token)

    edge = Edge(
        subject=entity,
        object=old_entity,
        subject_representation=e_representation,
        object_representation=o_representation,
    )

    return NodeDeepening(
        id=id,
        about_edge=edge,
        old_value=old_entity,
        new_value=new_entity,
        old_object_type=o_representation,
        new_object_type=n_representation,
    )


def parse_move(tree, id):
    subject_token = extract(tree, "entity_subject")
    predicate_token = extract(tree, "entity_predicate")
    object_token = extract(tree, "entity_object")

    subject, s_representation = get_entity_representation(subject_token)
    predicate, p_representation = get_entity_representation(predicate_token)
    object, o_representation = get_entity_representation(object_token)

    old_token = extract(tree, "old_entity")
    new_token = extract(tree, "new_entity")

    old, old_representation = get_entity_representation(old_token)
    new, new_representation = get_entity_representation(new_token)

    edge = Edge(
        subject=subject,
        predicate=predicate,
        object=object,
        subject_representation=s_representation,
        predicate_representation=p_representation,
        object_representation=o_representation,
    )

    return NodeMove(
        id=id,
        about_edge=edge,
        old_value=old,
        new_value=new,
        old_object_type=old_representation,
        new_object_type=new_representation,
    )


def parse_delete(tree, id):
    entity_token = extract(tree, "entity")
    entity, representation = get_entity_representation(entity_token)

    return NodeDeletion(
        id=id, about_node=entity, about_node_representation=representation
    )


def parse_unobsolete(tree, id):
    entity_token = extract(tree, "entity")
    entity, representation = get_entity_representation(entity_token)

    return NodeUnobsoletion(
        id=id, about_node=entity, about_node_representation=representation
    )


def parse_obsolete(tree, id):
    entity_token = extract(tree, "entity")
    entity, representation = get_entity_representation(entity_token)

    replacement_token = extract(tree, "replacement")

    return NodeObsoletion(
        id=id,
        about_node=entity,
        about_node_representation=representation,
        has_direct_replacement=replacement_token,
    )


def parse_rename(tree, id):
    old_token = extract(tree, "old_label")
    new_token = extract(tree, "new_label")
    old_language = extract(tree, "old_language")
    new_language = extract(tree, "new_language")

    # TODO old_token and new_token are enclosed in ''

    term_id_token = extract(tree, "id")
    if term_id_token is not None:
        entity, representation = get_entity_representation(term_id_token)
    else:
        entity = None
        representation = None

    return NodeRename(
        id=id,
        about_node=entity,
        about_node_representation=representation,
        old_value=old_token,
        new_value=new_token,
        old_language=old_language,
        new_language=new_language,
    )


def extract(tree, data):
    node = get_next(tree.find_data(data))
    if node is not None:
        node_token = next(get_tokens(node))
        return node_token
    else:
        return None


def get_tokens(tree):
    return tree.scan_values(lambda v: isinstance(v, Token))


def get_next(generator):
    try:
        res = next(generator)
        return res
    except StopIteration:
        return None


def get_entity_representation(entity):
    first_character = entity[0]
    last_character = entity[-1:]
    if first_character == "<" and last_character == ">":
        return entity, "uri"  # not removing brackets (TODO why?)
    if first_character == "'" and last_character == "'" and entity[1] != "'":
        return entity[1:-1], "label"
    if first_character == '"' and last_character == '"':
        return entity[1:-1], "literal"
    if entity[0:3] == '"""' and entity[-3:] == '"""':
        return entity[3:-3], "literal"
    if entity[0:3] == "'''" and entity[-3:] == "'''":
        return entity[3:-3], "literal"

    # TODO: use predefined set of prefixes to identify CURIEs
    return entity, "curie"
    # return entity, "error"


if __name__ == "__main__":
    #
    # MANUAL TESTING
    #
    example1 = "renamea 'abnormal ear' to 'abnormal ear morphology'"
    tree = kgcl_parser.parse(example1)
    print(tree)
    old = next(tree.find_data("old_label"))
    print(old)
    old_token = next(get_tokens(old))
    print(old_token)



import rdflib
from rdflib.namespace import (
    RDFS,
    RDF,
    OWL,
)
from rdflib import BNode, URIRef, Literal
from model.kgcl import (
    NodeRename,
    NodeObsoletion,
    NodeUnobsoletion,
    NodeDeletion,
    NodeMove,
    NodeDeepening,
    NodeShallowing,
    EdgeCreation,
    EdgeDeletion,
    PredicateChange,
    NodeCreation,
    ClassCreation,
    PlaceUnder,
    RemoveUnder,
    NewSynonym,
    RemovedNodeFromSubset,
    ExistentialRestrictionCreation,
    ExistentialRestrictionDeletion,
)
from diff.change_detection import (
    detect_renamings,
    detect_node_moves,
    detect_predicate_changes,
)
from diff.owlstar_sublanguage import (
    get_triple_annotations,
    get_bnodes_2_triple_annotations,
)
from diff.graph_diff import (
    get_added_triple_annotations,
    get_deleted_triple_annotations,
    get_added_existentials,
    get_deleted_existentials,
)
from diff.render_operations import render
from model.ontology_model import Edge, Annotation


def id_generator():
    id = 0
    while True:
        yield id
        id += 1


id_gen = id_generator()


class ExistentialChangeSummary:
    """
    Dataclass holding information about (atomic) existential restriction changes.
    """

    def __init__(self):

        self.existential_additions = []
        self.existential_deletions = []
        self.covered_triples_triple_existential_additions = []
        self.covered_triples_triple_existential_deletions = []

    def get_commands(self):
        kgcl_commands = []
        for k in self.existential_additions:
            kgcl_commands.append(k)
        for k in self.existential_deletions:
            kgcl_commands.append(k)
        return kgcl_commands

    def get_existential_additions(self):
        return self.existential_additions

    def get_existential_deletions(self):
        return self.existential_deletions

    def get_summary_KGCL_commands(self):
        out = (
            "ExistentialRestrictionCreations: "
            + str(len(self.existential_additions))
            + "\n"
            "ExistentialRestrictionDeletions: "
            + str(len(self.existential_deletions))
            + "\n"
        )
        return out

    def get_summary_RDF_triples(self):
        out = (
            "ExistentialRestrictionCreations"
            + str(len(self.covered_triples_triple_existential_additions))
            + "\n"
            "ExistentialRestrictionDeletions"
            + str(len(self.covered_triples_triple_existential_deletions))
            + "\n"
        )
        return out

    # RDF data

    def get_covered_triples_existential_additions(self):
        return self.covered_triples_triple_existential_additions

    def get_covered_triples_existential_deletions(self):
        return self.covered_triples_triple_existential_deletions

    def add_covered_triples_existential_additions(self, triples):
        for t in triples:
            self.covered_triples_triple_existential_additions.append(t)

    def add_covered_triples_existential_deletions(self, triples):
        for t in triples:
            self.covered_triples_triple_existential_deletions.append(t)

    # KGCL data

    def add_existential_addition(self, i):
        self.existential_additions.append(i)

    def add_existential_deletion(self, i):
        self.existential_deletions.append(i)


def generate_atomic_existential_commands(g1, g2):
    """
    Given two graphs g1 and g2,
    return all ExistentialRestrictionCreation and
    ExistentialRestrictionDeletion to account for their diff.
    """
    summary = ExistentialChangeSummary()

    added = get_added_existentials(g1, g2)
    deleted = get_deleted_existentials(g1, g2)

    existential_additions, covered = generate_existential_additions(added)
    summary.add_covered_triples_existential_additions(covered)

    existential_deletions, covered = generate_existential_deletions(deleted)
    summary.add_covered_triples_existential_deletions(covered)

    for a in existential_additions:
        summary.add_existential_addition(render(a))
    for d in existential_deletions:
        summary.add_existential_deletion(render(d))

    return summary


def generate_existential_deletions(deleted):
    """
    Return ExistentialRestrictionDeletion instances for given (deleted) triples.
    """
    covered = rdflib.Graph()
    kgcl = []

    for a in deleted:
        subclass = str(a.subclass)
        property = str(a.property)
        filler = str(a.filler)

        id = "test_id_" + str(next(id_gen))

        node = ExistentialRestrictionDeletion(
            id=id, subclass=subclass, property=property, filler=filler
        )

        kgcl.append(node)
        for t in a.triples:
            covered.add(t)

    return kgcl, covered


def generate_existential_additions(added):
    """
    Return ExistentialRestrictionCreation instances for given (added) triples.
    """
    covered = rdflib.Graph()
    kgcl = []

    for a in added:
        subclass = str(a.subclass)
        property = str(a.property)
        filler = str(a.filler)

        id = "test_id_" + str(next(id_gen))

        node = ExistentialRestrictionCreation(
            id=id, subclass=subclass, property=property, filler=filler
        )

        kgcl.append(node)
        for t in a.triples:
            covered.add(t)

    return kgcl, covered



import rdflib
from rdflib.namespace import (
    RDFS,
    OWL,
    RDF,
)
from rdflib import BNode, URIRef
import time


class TripleAnnotation:
    # TODO: store blank node associated with the annotation
    def __init__(self, s, p, t, ap, a):
        self.source = s
        self.property = p
        self.target = t
        self.annotation_property = ap
        self.annotation = a
        self.triples = set()

    # TODO check that all triples have the same blank node
    def add_triple(self, t):
        self.triples.add(t)

    def add_triples(self, ts):
        self.triples.update(ts)

    def __repr__(self):
        return (
            "<<"
            + str(self.source)
            + " "
            + str(self.property)
            + " "
            + str(self.target)
            + ">> "
            + str(self.annotation_property)
            + " "
            + str(self.annotation)
        )

    def __eq__(self, other):
        if isinstance(other, TripleAnnotation):
            return (
                self.source == other.source
                and self.property == other.property
                and self.target == other.target
                and self.annotation_property == other.annotation_property
                and self.annotation == other.annotation
            )
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __key(self):
        return (
            self.source,
            self.property,
            self.target,
            self.annotation_property,
            self.annotation,
        )

    def __hash__(self):
        return hash(self.__key())


class ExistentialRestriction:
    # TODO: store blank node associated with the existential restriction
    def __init__(self, sub, prop, fil):
        self.subclass = sub
        self.property = prop
        self.filler = fil
        self.triples = set()

    def add_triple(self, t):
        self.triples.add(t)

    def add_triples(self, ts):
        self.triples.update(ts)

    def __repr__(self):
        return (
            self.subclass
            + " "
            + "SubClassOf"
            + " "
            + self.property
            + " "
            + "some"
            + " "
            + self.filler
        )

    def __eq__(self, other):
        if isinstance(other, ExistentialRestriction):
            return (
                self.subclass == other.subclass
                and self.property == other.property
                and self.filler == other.filler
            )
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __key(self):
        return (
            self.subclass,
            self.property,
            self.filler,
        )

    def __hash__(self):
        return hash(self.__key())


def get_thin_triples(g):
    res = rdflib.Graph()
    for s, p, o in g.triples((None, None, None)):
        if (
            not isinstance(s, BNode)
            and not isinstance(o, BNode)
            # and p != RDFS.subClassOf
        ):
            res.add((s, p, o))

    return res


def get_atomic_subsumptions(g):
    res = rdflib.Graph()
    for s, p, o in g.triples((None, RDFS.subClassOf, None)):
        if not isinstance(s, BNode) and not isinstance(o, BNode):
            res.add((s, p, o))

    return res


def get_atomic_existentials(g):
    some_values_from = set(g.subjects(predicate=OWL.someValuesFrom))
    on_property = set(g.subjects(predicate=OWL.onProperty))

    intersection = some_values_from & on_property

    existential_2_classes = {}  # map from blank node to (named) subclasses
    for i in intersection:
        existential_2_classes[i] = []
        if isinstance(i, BNode):  # this check should be unnecessary
            for s, p, o in g.triples((None, RDFS.subClassOf, i)):
                if not isinstance(s, BNode):
                    existential_2_classes[i].append(str(s))

    owlstar_axiom = []
    bnodes_2_existentials = get_bnodes_2_atomic_existentials(g)
    for i, subclasses in existential_2_classes.items():
        # bnode = str(i)  # don't really need the bnode
        filler = ""
        bnode_filler = True
        property = ""
        bnode_property = True

        # get property
        for s, p, o in g.triples((i, OWL.onProperty, None)):
            if not isinstance(o, BNode):  # this test is unnecessary
                property = str(o)
                bnode_property = False

        # get filler
        for s, p, o in g.triples((i, OWL.someValuesFrom, None)):
            if not isinstance(o, BNode):
                filler = str(o)
                bnode_filler = False

        if not bnode_property and not bnode_filler:
            for subclass in subclasses:
                ex = ExistentialRestriction(subclass, property, filler)
                ex.add_triples(bnodes_2_existentials[i])
                owlstar_axiom.append(ex)
                # owlstar_axiom.append((subclass, property, filler))

    return owlstar_axiom


def get_bnodes_2_atomic_existentials(g):
    some_values_from = set(g.subjects(predicate=OWL.someValuesFrom))
    on_property = set(g.subjects(predicate=OWL.onProperty))
    # restriction = set(g.subjects(predicate=rdf.type)) # Restriction

    intersection = some_values_from & on_property

    existential_2_classes = {}  # map from blank node to (named) subclass
    for i in intersection:
        existential_2_classes[i] = []
        if isinstance(i, BNode):  # this check should be unnecessary
            for s, p, o in g.triples((None, RDFS.subClassOf, i)):
                if not isinstance(s, BNode):
                    existential_2_classes[i].append(s)

    bnode_triples = {}
    for i, subclasses in existential_2_classes.items():
        bnode_triples[i] = []

        filler = i
        bnode_filler = True
        property = i
        bnode_property = True

        # get property
        for s, p, o in g.triples((i, OWL.onProperty, None)):
            if not isinstance(o, BNode):
                property = o
                bnode_property = False

        # get filler
        for s, p, o in g.triples((i, OWL.someValuesFrom, None)):
            if not isinstance(o, BNode):
                filler = o
                bnode_filler = False

        if not bnode_property and not bnode_filler:
            for subclass in subclasses:
                bnode_triples[i].append((i, OWL.some_values_from, filler))
                bnode_triples[i].append((i, OWL.onProperty, property))
                bnode_triples[i].append((subclass, RDFS.subClassOf, i))

    return bnode_triples


def get_triple_annotations(g):
    source = set(g.subjects(predicate=OWL.annotatedSource))
    property = set(g.subjects(predicate=OWL.annotatedProperty))
    target = set(g.subjects(predicate=OWL.annotatedTarget))

    intersection = source & property & target

    exclude = {
        OWL.annotatedSource,
        OWL.annotatedProperty,
        OWL.annotatedTarget,
        RDF.type,
    }

    bnodes_2_triple_annotations = get_bnodes_2_triple_annotations(g)
    annotations = []
    for i in intersection:
        if isinstance(i, BNode):  # this check should be unnecessary
            # NB these generators are singletons
            source = next(g.objects(subject=i, predicate=OWL.annotatedSource))
            property = next(g.objects(subject=i, predicate=OWL.annotatedProperty))
            target = next(g.objects(subject=i, predicate=OWL.annotatedTarget))
            for s, p, o in g.triples((i, None, None)):
                if (
                    p not in exclude
                    and not isinstance(o, BNode)
                    and not isinstance(p, BNode)
                    and not isinstance(source, BNode)
                    and not isinstance(property, BNode)
                    and not isinstance(target, BNode)
                ):
                    ta = TripleAnnotation(source, property, target, p, o)
                    ta.add_triples(bnodes_2_triple_annotations[s])
                    annotations.append(ta)

                    # annotations.append((source, property, target, p, o))
                    # (str(source), str(property), str(target), str(p), str(o))

    return annotations


def get_bnodes_2_triple_annotations(g):
    source = set(g.subjects(predicate=OWL.annotatedSource))
    property = set(g.subjects(predicate=OWL.annotatedProperty))
    target = set(g.subjects(predicate=OWL.annotatedTarget))

    # get blank nodes that have all three required predicates
    intersection = source & property & target

    exclude = {
        OWL.annotatedSource,
        OWL.annotatedProperty,
        OWL.annotatedTarget,
        # RDF.type,
    }

    annotations = {}
    for i in intersection:
        annotations[i] = []
        if isinstance(i, BNode):  # this check should be unnecessary
            # NB these generators are singletons
            source = next(g.objects(subject=i, predicate=OWL.annotatedSource))
            property = next(g.objects(subject=i, predicate=OWL.annotatedProperty))
            target = next(g.objects(subject=i, predicate=OWL.annotatedTarget))

            annotations[i].append((i, OWL.annotatedSource, source))
            annotations[i].append((i, OWL.annotatedProperty, property))
            annotations[i].append((i, OWL.annotatedTarget, target))

            for s, p, o in g.triples((i, None, None)):
                if (
                    p not in exclude
                    and not isinstance(o, BNode)
                    and not isinstance(p, BNode)
                    and not isinstance(source, BNode)
                    and not isinstance(property, BNode)
                    and not isinstance(target, BNode)
                ):
                    annotations[i].append((s, p, o))

    return annotations


def render_triple_annotation(a):
    return (
        "<<"
        # + a[0].n3()
        + a[0]
        + " "
        # + a[1].n3()
        + a[1]
        + " "
        # + a[2].n3()
        + a[2]
        + ">> "
        # + a[3].n3()
        + a[3]
        + " "
        # + a[4].n3()
        + a[4]
    )


def render_atomic_existential(a):
    return (
        "<<"
        + a[0]
        + " "
        + a[1]
        + " "
        + a[2]
        + ">> "
        + "os:interpretation"
        + " "
        + "os:AllSomeInterpretation"
    )


if __name__ == "__main__":
    g = rdflib.Graph()

    start = time.time()
    g.load("n3/obi/obi_1.nt", format="nt")
    # g.load("n3/uberon/uberon_1.nt", format="nt")
    end = time.time()
    print(end - start)

    start = time.time()
    thin = get_thin_triples(g)
    end = time.time()
    print(end - start)

    start = time.time()
    axioms = get_atomic_existentials(g)
    end = time.time()
    print(end - start)

    print(len(axioms))

    annotations = get_triple_annotations(g)
    print(len(annotations))
    # for a in annotations:
    #    print(str(a[0]) + " " + str(a[1]))
    # for a in axioms:
    # print(a)



import rdflib
from rdflib.namespace import (
    RDFS,
    RDF,
    OWL,
)
from rdflib import BNode, URIRef, Literal
from model.kgcl import (
    NodeRename,
    NodeObsoletion,
    NodeUnobsoletion,
    NodeDeletion,
    NodeMove,
    NodeDeepening,
    NodeShallowing,
    EdgeCreation,
    EdgeDeletion,
    PredicateChange,
    NodeCreation,
    ClassCreation,
    PlaceUnder,
    RemoveUnder,
    NewSynonym,
    RemovedNodeFromSubset,
    ExistentialRestrictionCreation,
    ExistentialRestrictionDeletion,
)
from diff.change_detection import (
    detect_renamings,
    detect_node_moves,
    detect_predicate_changes,
)
from diff.owlstar_sublanguage import (
    get_triple_annotations,
    get_bnodes_2_triple_annotations,
)
from diff.graph_diff import (
    get_added_triple_annotations,
    get_deleted_triple_annotations,
    get_added_existentials,
    get_deleted_existentials,
)
from diff.render_operations import render
from model.ontology_model import Edge, Annotation


def id_generator():
    id = 0
    while True:
        yield id
        id += 1


id_gen = id_generator()


class TripleAnnotationChangeSummary:
    """
    Dataclass holding information about triple annotation axioms.
    """

    def __init__(self):

        self.triple_annotations_additions = []
        self.triple_annotations_deletions = []
        self.covered_triples_triple_annotations_additions = []
        self.covered_triples_triple_annotations_deletions = []

    def get_commands(self):
        kgcl_commands = []
        for k in self.triple_annotations_additions:
            kgcl_commands.append(k)
        for k in self.triple_annotations_deletions:
            kgcl_commands.append(k)
        return kgcl_commands

    def get_triple_annotation_additions(self):
        return self.triple_annotations_additions

    def get_triple_annotation_deletions(self):
        return self.triple_annotations_deletions

    def get_summary_KGCL_commands(self):
        out = (
            # "Triple Annotation Additions: "
            "EdgeCreations (with annotation): "
            + str(len(self.triple_annotations_additions))
            + "\n"
            # "Triple Annotation Deletions: "
            + "EdgeDeletion (with annotation): "
            + str(len(self.triple_annotations_deletions))
            + "\n"
        )
        return out

    def get_summary_RDF_triples(self):
        out = (
            # "Triple Annotation Additions: "
            "EdgeCreations (with annotation): "
            + str(len(self.covered_triples_triple_annotations_additions))
            + "\n"
            # + "Triple Annotation Deletions: "
            + "EdgeDeletion (with annotation): "
            + str(len(self.covered_triples_triple_annotations_deletions))
            + "\n"
        )
        return out

    # RDF data

    def get_covered_triples_annotation_additions(self):
        return self.covered_triples_triple_annotations_additions

    def get_covered_triples_annotation_deletions(self):
        return self.covered_triples_triple_annotations_deletions

    def add_covered_triples_annotation_additions(self, triples):
        for t in triples:
            self.covered_triples_triple_annotations_additions.append(t)

    def add_covered_triples_annotation_deletions(self, triples):
        for t in triples:
            self.covered_triples_triple_annotations_deletions.append(t)

    # KGCL data

    def add_triple_annotation_addition(self, i):
        self.triple_annotations_additions.append(i)

    def add_triple_annotation_deletion(self, i):
        self.triple_annotations_deletions.append(i)


def generate_triple_annotation_commands(g1, g2):
    """
    Given two graphs g1 and g2,
    return all (annotated) EdgeCreation and
    (annotated) EdgeDeletions to account for their diff.
    """
    summary = TripleAnnotationChangeSummary()

    added = get_added_triple_annotations(g1, g2)
    deleted = get_deleted_triple_annotations(g1, g2)

    additions, covered = generate_triple_annotation_additions(added)
    summary.add_covered_triples_annotation_additions(covered)

    deletions, covered = generate_triple_annotation_deletions(deleted)
    summary.add_covered_triples_annotation_deletions(covered)

    for a in additions:
        summary.add_triple_annotation_addition(render(a))
    for d in deletions:
        summary.add_triple_annotation_deletion(render(d))

    return summary


def generate_triple_annotation_additions(added):
    """
    Return EdgeCreation instances for given (added) triples.
    """
    covered = rdflib.Graph()
    kgcl = []

    for a in added:
        source = str(a.source)
        property = str(a.property)
        target = str(a.target)
        target_type = get_type(a.target)
        annotation_property = str(a.annotation_property)
        annotation = str(a.annotation)
        annotation_type = get_type(a.annotation)

        id = "test_id_" + str(next(id_gen))

        language = get_language_tag(a.target)
        datatype = get_datatype(a.target)
        if datatype is not None:
            datatype = "<" + datatype + ">"  # expect data types without curies

        annotation = Annotation(
            property=annotation_property, filler=annotation, filler_type=annotation_type
        )

        node = EdgeCreation(
            id=id,
            subject=source,
            predicate=property,
            object=target,
            object_type=target_type,
            annotation_set=annotation,
            language=language,
            datatype=datatype,
        )

        kgcl.append(node)
        for t in a.triples:
            covered.add(t)

    return kgcl, covered


def generate_triple_annotation_deletions(deleted):
    """
    Return EdgeDeletion instances for given (deleted) triples.
    """
    covered = rdflib.Graph()
    kgcl = []

    for a in deleted:
        source = str(a.source)
        property = str(a.property)
        target = str(a.target)
        target_type = get_type(a.target)
        annotation_property = str(a.annotation_property)
        annotation = str(a.annotation)
        annotation_type = get_type(a.annotation)

        id = "test_id_" + str(next(id_gen))

        language = get_language_tag(a.target)
        datatype = get_datatype(a.target)
        if datatype is not None:
            datatype = "<" + datatype + ">"  # expect data types without curies

        annotation = Annotation(
            property=annotation_property, filler=annotation, filler_type=annotation_type
        )

        node = EdgeDeletion(
            id=id,
            subject=source,
            predicate=property,
            object=target,
            object_type=target_type,
            annotation_set=annotation,
            language=language,
            datatype=datatype,
        )

        kgcl.append(node)
        for t in a.triples:
            covered.add(t)

    return kgcl, covered


# TODO: factor these out in some kind of utility library


def get_type(rdf_entity):
    if isinstance(rdf_entity, URIRef):
        return "uri"
    elif isinstance(rdf_entity, Literal):
        return "literal"
    else:
        return "Error"


def get_language_tag(rdf_entity):
    if isinstance(rdf_entity, Literal):
        return rdf_entity.language
    else:
        return None


def get_datatype(rdf_entity):
    if isinstance(rdf_entity, Literal) and rdf_entity.datatype is not None:
        return str(rdf_entity.datatype)
    else:
        return None



import rdflib
from rdflib.namespace import (
    RDFS,
    RDF,
    OWL,
)
from rdflib import BNode, URIRef, Literal
from model.kgcl import (
    NodeRename,
    NodeObsoletion,
    NodeUnobsoletion,
    NodeDeletion,
    NodeMove,
    NodeDeepening,
    NodeShallowing,
    EdgeCreation,
    EdgeDeletion,
    PredicateChange,
    NodeCreation,
    ClassCreation,
    NewSynonym,
    PlaceUnder,
    RemoveUnder,
    RemovedNodeFromSubset,
    ExistentialRestrictionCreation,
    ExistentialRestrictionDeletion,
)
from model.ontology_model import Edge
import grammar.parser


# TODO: maintain this dictionary in a file
prefix_2_uri = {
    "obo": "http://purl.obolibrary.org/obo/",
    "ex": "http://example.org/",
    "oboInOwl": "http://www.geneontology.org/formats/oboInOwl#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "xsd": "http://www.w3.org/2001/XMLSchema#",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "owl": "http://www.w3.org/2002/07/owl#",
    "swrl": "http://www.w3.org/2003/11/swrl#",
    "oio": "http://www.geneontology.org/formats/oboInOwl#",
    "dce": "http://purl.org/dc/elements/1.1/",
    "dct": "http://purl.org/dc/terms/",
    "foaf": "http://xmlns.com/foaf/0.1/",
    "protege": "http://protege.stanford.edu/plugins/owl/protege#",
    "BFO": "http://purl.obolibrary.org/obo/BFO_",
    "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
    "CL": "http://purl.obolibrary.org/obo/CL_",
    "IAO": "http://purl.obolibrary.org/obo/IAO_",
    "NCBITaxon": "http://purl.obolibrary.org/obo/NCBITaxon_",
    "OBI": "http://purl.obolibrary.org/obo/OBI_",
    "PR": "http://purl.obolibrary.org/obo/PR_",
    "obo": "http://purl.obolibrary.org/obo/",
    "UP": "http://purl.uniprot.org/uniprot/",
    "UC": "http://purl.uniprot.org/core/",
    "PRO": "http://www.uniprot.org/annotation/PRO_",
    "faldo": "http://biohackathon.org/resource/faldo#",
}


def get_labels(graph):
    """
    Returns a map from IRIs in a graph to (a set of) labels.
    """
    entity_2_label = {}
    for s, p, o in graph.triples((None, RDFS.label, None)):
        ss = str(s)
        os = str(o)
        if "'" not in os:
            if ss not in entity_2_label:
                entity_2_label[ss] = []
            entity_2_label[ss].append(os)

    return entity_2_label


def render_instances(kgcl, graph):
    """
    Takes a KGCL patch for a graph and
    returns a more readable patch
    in which IRI's are replaced by CURIEs and labels where possible.
    """
    labelling = get_labels(graph)

    pretty_print_kgcl = []
    for k in kgcl:
        # print(k)
        kgcl_instance = grammar.parser.parse_statement(k)
        # print(render_instance(kgcl_instance, labelling))  # pretty print
        # render_instance(kgcl_instance, labelling)
        pretty_print_kgcl.append(render_instance(kgcl_instance, labelling))

    return pretty_print_kgcl


def has_label(entity, labelling):
    """
    Returns the label for an entity if it exists
    and the entity itself otherwise.
    """
    if entity in labelling:
        return "'" + labelling[entity][0] + "'"
    else:
        return entity


def curie_entity(entity):
    """
    Returns the CURIE for an entity if it exists
    and the entity itself otherwise.
    """
    for prefix, curie in prefix_2_uri.items():
        if curie in entity:
            return entity.replace(curie, prefix + ":")[1:-1]
    return entity


def render_entity(entity, type, labelling):
    """
    Returns an encoding of the given entity using either
    a CURIE, a label, a literal, or a URI.
    """
    entity = str(entity)
    entity = repr(entity)[1:-1]
    if type == "uri":
        labelling = has_label(entity[1:-1], labelling)
        if entity[1:-1] != labelling:
            return labelling

        curing = curie_entity(entity)
        if entity != curing:
            return curing
        # return "<" + entity + ">"
        return entity

    elif type == "label":
        if "'" not in entity:
            return "'" + entity + "'"

    elif type == "literal":
        if '"' not in entity:
            return '"' + entity + '"'
        elif "'''" not in entity and entity[-1] != "'":
            return "'''" + entity + "'''"
        elif '"""' not in entity and entity[-1] != '"':
            return '"""' + entity + '"""'
        else:
            # return "Error  " + entity
            print("Rendering error: " + entity)
            raise
    else:
        # return "Error  " + entity
        print("Rendering error: " + entity)
        raise


def render_instance(kgclInstance, labelling):
    if type(kgclInstance) is NodeRename:
        # TODO: subject could be 'None'? (not in KGCL Diff)
        # passing an empty labelling because we do not want to repeat labels
        subject = render_entity(kgclInstance.about_node, "uri", {})
        old = kgclInstance.old_value
        new = kgclInstance.new_value

        new_language = kgclInstance.new_language
        old_language = kgclInstance.old_language

        if old_language is not None:
            old = old + "@" + old_language

        if new_language is not None:
            new = new + "@" + new_language

        return "rename " + subject + " from " + old + " to " + new

    if type(kgclInstance) is NodeObsoletion:
        subject = render_entity(kgclInstance.about_node, "uri", labelling)
        replacement = render_entity(
            kgclInstance.has_direct_replacement, "uri", labelling
        )
        if kgclInstance.has_direct_replacement is not None:
            return "obsolete " + subject + " with replacement " + replacement
        else:
            return "obsolete " + subject

    if type(kgclInstance) is NodeUnobsoletion:
        subject = render_entity(kgclInstance.about_node, "uri", labelling)
        return "unobsolete " + subject

    if type(kgclInstance) is NodeDeletion:
        subject = render_entity(kgclInstance.about_node, "uri", labelling)
        return "delete " + subject

    if type(kgclInstance) is NodeMove:
        subject = render_entity(kgclInstance.about_edge.subject, "uri", labelling)
        predicate = render_entity(kgclInstance.about_edge.predicate, "uri", labelling)
        new = render_entity(
            kgclInstance.new_value, kgclInstance.new_object_type, labelling
        )
        old = render_entity(
            kgclInstance.old_value, kgclInstance.old_object_type, labelling
        )

        return (
            "move "
            + subject
            + " "
            + predicate
            + " "
            + old
            + " "
            + "from"
            + " "
            + old
            + " "
            + "to"
            + " "
            + new
        )

    if type(kgclInstance) is EdgeCreation:
        if kgclInstance.annotation_set is None:
            return render_edge_creation(kgclInstance, labelling)
        else:
            return render_annotation_creation(kgclInstance, labelling)

    if type(kgclInstance) is EdgeDeletion:
        if kgclInstance.annotation_set is None:
            return render_edge_deletion(kgclInstance, labelling)
        else:
            return render_annotation_deletion(kgclInstance, labelling)

    if type(kgclInstance) is PredicateChange:
        subject = render_entity(kgclInstance.about_edge.subject, "uri", labelling)
        object = render_entity(
            kgclInstance.about_edge.object,
            kgclInstance.about_edge.object_representation,
            labelling,
        )

        new = render_entity(
            kgclInstance.new_value,
            "uri",
            labelling,
        )
        old = render_entity(kgclInstance.old_value, "uri", labelling)

        if kgclInstance.language is not None:
            object += "@" + kgclInstance.language

        if kgclInstance.datatype is not None:
            object += "^^" + kgclInstance.datatype

        return (
            "change relationship between "
            + subject
            + " and "
            + object
            + " from "
            + old
            + " to "
            + new
        )

    if type(kgclInstance) is NodeCreation:
        subject = render_entity(kgclInstance.about_node, "uri", labelling)
        label = render_entity(kgclInstance.name, "label", labelling)
        if kgclInstance.name is not None:
            return "create node " + subject + " " + label
        else:
            return "create " + subject

    if type(kgclInstance) is ClassCreation:
        subject = render_entity(kgclInstance.node_id, "uri", labelling)
        return "create " + subject

    if type(kgclInstance) is NewSynonym:
        subject = render_entity(kgclInstance.about_node, "uri", labelling)
        synonym = render_entity(kgclInstance.new_value, "label", labelling)
        qualifier = kgclInstance.qualifier
        language = kgclInstance.language

        if language is not None:
            synonym = synonym + "@" + language

        if qualifier is not None:
            return (
                "create " + qualifier + " synonym" + " " + synonym + " for " + subject
            )
        else:
            return "create synonym " + synonym + " for " + subject

    if type(kgclInstance) is ExistentialRestrictionCreation:
        subclass = render_entity(kgclInstance.subclass, "uri", labelling)
        property = render_entity(kgclInstance.property, "uri", labelling)
        filler = render_entity(kgclInstance.filler, "uri", labelling)
        return "add " + subclass + " SubClassOf " + property + " some " + filler

    if type(kgclInstance) is ExistentialRestrictionDeletion:
        subclass = render_entity(kgclInstance.subclass, "uri", labelling)
        property = render_entity(kgclInstance.property, "uri", labelling)
        filler = render_entity(kgclInstance.filler, "uri", labelling)
        return "delete " + subclass + " SubClassOf " + property + " some " + filler

    if type(kgclInstance) is PlaceUnder:
        subclass = render_entity(kgclInstance.subject, "uri", labelling)
        superclass = render_entity(kgclInstance.object, "uri", labelling)
        return "add " + subclass + " SubClassOf " + superclass

    if type(kgclInstance) is RemoveUnder:
        subclass = render_entity(kgclInstance.subject, "uri", labelling)
        superclass = render_entity(kgclInstance.object, "uri", labelling)
        return "delete " + subclass + " SubClassOf " + superclass


def render_annotation_creation(kgclInstance, labelling):
    subject = render_entity(kgclInstance.subject, "uri", labelling)
    predicate = render_entity(kgclInstance.predicate, "uri", labelling)
    object = render_entity(kgclInstance.object, kgclInstance.object_type, labelling)

    annotation = kgclInstance.annotation_set
    annotation_property = render_entity(annotation.property, "uri", labelling)
    annotation_filler = render_entity(
        annotation.filler, annotation.filler_type, labelling
    )

    language = kgclInstance.language
    datatype = kgclInstance.datatype

    if language is not None:
        object = object + "@" + language
    if datatype is not None:
        object = object + "^^" + datatype

    return (
        "create edge <<"
        + subject
        + " "
        + predicate
        + " "
        + object
        + ">>"
        + " "
        + annotation_property
        + " "
        + annotation_filler
    )


def render_annotation_deletion(kgclInstance, labelling):
    subject = render_entity(kgclInstance.subject, "uri", labelling)
    predicate = render_entity(kgclInstance.predicate, "uri", labelling)
    object = render_entity(kgclInstance.object, kgclInstance.object_type, labelling)

    annotation = kgclInstance.annotation_set
    annotation_property = render_entity(annotation.property, "uri", labelling)
    annotation_filler = render_entity(
        annotation.filler, annotation.filler_type, labelling
    )

    language = kgclInstance.language
    datatype = kgclInstance.datatype

    if language is not None:
        object = object + "@" + language
    if datatype is not None:
        object = object + "^^" + datatype

    return (
        "delete edge <<"
        + subject
        + " "
        + predicate
        + " "
        + object
        + ">>"
        + " "
        + annotation_property
        + " "
        + annotation_filler
    )


def render_edge_deletion(kgclInstance, labelling):
    subject = render_entity(kgclInstance.subject, "uri", labelling)
    predicate = render_entity(kgclInstance.predicate, "uri", labelling)
    object = render_entity(kgclInstance.object, kgclInstance.object_type, labelling)

    language = kgclInstance.language
    datatype = kgclInstance.datatype

    base = "delete edge " + subject + " " + predicate + " " + object

    if language is not None:
        return base + "@" + language
    elif datatype is not None:
        return base + "^^" + datatype
    else:
        return base


def render_edge_creation(kgclInstance, labelling):
    subject = render_entity(kgclInstance.subject, "uri", labelling)
    predicate = render_entity(kgclInstance.predicate, "uri", labelling)
    object = render_entity(kgclInstance.object, kgclInstance.object_type, labelling)

    language = kgclInstance.language
    datatype = kgclInstance.datatype

    base = "create edge " + subject + " " + predicate + " " + object

    if language is not None:
        return base + "@" + language
    elif datatype is not None:
        return base + "^^" + datatype
    else:
        return base



# Auto generated from basics.yaml by pythongen.py version: 0.9.0
# Generation date: 2021-05-11 09:08
# Schema: basics
#
# id: https://w3id.org/kgcl/basics
# description: Core predicates
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_model.types import String

metamodel_version = "1.7.0"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
BASICS = CurieNamespace('basics', 'https://w3id.org/kgcl/basics/')
DCTERMS = CurieNamespace('dcterms', 'http://purl.org/dc/terms/')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
DEFAULT_ = BASICS


# Types

# Class references




# Enumerations


# Slots
class slots:
    pass

slots.id = Slot(uri=BASICS.id, name="id", curie=BASICS.curie('id'),
                   model_uri=BASICS.id, domain=None, range=URIRef)

slots.description = Slot(uri=DCTERMS.description, name="description", curie=DCTERMS.curie('description'),
                   model_uri=BASICS.description, domain=None, range=Optional[str])



# Auto generated from ontology_model.yaml by pythongen.py version: 0.9.0
# Generation date: 2021-09-23 19:44
# Schema: kgcl-ontology-model
#
# id: https://w3id.org/kgcl/ontology
# description: A basic bare-bones model of an ontology or ontology-like structure. The purpose is not to provide a
#              complete model, rather just sufficient structure for domain and range constraints in the ocl model
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import String

metamodel_version = "1.7.0"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
DCTERMS = CurieNamespace('dcterms', 'http://purl.org/dc/terms/')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
OIO = CurieNamespace('oio', 'http://www.geneontology.org/formats/oboInOwl#')
OM = CurieNamespace('om', 'http://w3id.org/kgcl/om/')
OWL = CurieNamespace('owl', 'http://www.w3.org/2002/07/owl#')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
DEFAULT_ = OM


# Types

# Class references
class NodeId(extended_str):
    pass


class ClassNodeId(NodeId):
    pass


class InstanceNodeId(NodeId):
    pass


class OntologyElement(YAMLRoot):
    """
    Any component of an ontology or knowledge graph
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.OntologyElement
    class_class_curie: ClassVar[str] = "om:OntologyElement"
    class_name: ClassVar[str] = "ontology element"
    class_model_uri: ClassVar[URIRef] = OM.OntologyElement


@dataclass
class PropertyValue(OntologyElement):
    """
    a property-value pair
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.PropertyValue
    class_class_curie: ClassVar[str] = "om:PropertyValue"
    class_name: ClassVar[str] = "property value"
    class_model_uri: ClassVar[URIRef] = OM.PropertyValue

    property: Optional[Union[str, NodeId]] = None
    filler: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.property is not None and not isinstance(self.property, NodeId):
            self.property = NodeId(self.property)

        if self.filler is not None and not isinstance(self.filler, str):
            self.filler = str(self.filler)

        super().__post_init__(**kwargs)


@dataclass
class Annotation(PropertyValue):
    """
    owl annotations. Not to be confused with annotations sensu GO
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.Annotation
    class_class_curie: ClassVar[str] = "om:Annotation"
    class_name: ClassVar[str] = "annotation"
    class_model_uri: ClassVar[URIRef] = OM.Annotation

    property: Optional[Union[str, NodeId]] = None
    filler: Optional[str] = None
    annotation_set: Optional[Union[dict, "Annotation"]] = None
    property_type: Optional[str] = None
    filler_type: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.property is not None and not isinstance(self.property, NodeId):
            self.property = NodeId(self.property)

        if self.filler is not None and not isinstance(self.filler, str):
            self.filler = str(self.filler)

        if self.annotation_set is not None and not isinstance(self.annotation_set, Annotation):
            self.annotation_set = Annotation(**as_dict(self.annotation_set))

        if self.property_type is not None and not isinstance(self.property_type, str):
            self.property_type = str(self.property_type)

        if self.filler_type is not None and not isinstance(self.filler_type, str):
            self.filler_type = str(self.filler_type)

        super().__post_init__(**kwargs)


@dataclass
class Node(OntologyElement):
    """
    Any named entity in an ontology. May be a class, individual, property
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.Node
    class_class_curie: ClassVar[str] = "om:Node"
    class_name: ClassVar[str] = "node"
    class_model_uri: ClassVar[URIRef] = OM.Node

    id: Union[str, NodeId] = None
    name: Optional[str] = None
    annotation_set: Optional[Union[dict, Annotation]] = None
    owl_type: Optional[Union[str, "OwlTypeEnum"]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, NodeId):
            self.id = NodeId(self.id)

        if self.name is not None and not isinstance(self.name, str):
            self.name = str(self.name)

        if self.annotation_set is not None and not isinstance(self.annotation_set, Annotation):
            self.annotation_set = Annotation(**as_dict(self.annotation_set))

        if self.owl_type is not None and not isinstance(self.owl_type, OwlTypeEnum):
            self.owl_type = OwlTypeEnum(self.owl_type)

        super().__post_init__(**kwargs)


@dataclass
class ClassNode(Node):
    """
    A node that is a class
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OWL.Class
    class_class_curie: ClassVar[str] = "owl:Class"
    class_name: ClassVar[str] = "class node"
    class_model_uri: ClassVar[URIRef] = OM.ClassNode

    id: Union[str, ClassNodeId] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, ClassNodeId):
            self.id = ClassNodeId(self.id)

        super().__post_init__(**kwargs)


@dataclass
class InstanceNode(Node):
    """
    A node that is an individual
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OWL.NamedIndividual
    class_class_curie: ClassVar[str] = "owl:NamedIndividual"
    class_name: ClassVar[str] = "instance node"
    class_model_uri: ClassVar[URIRef] = OM.InstanceNode

    id: Union[str, InstanceNodeId] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, InstanceNodeId):
            self.id = InstanceNodeId(self.id)

        super().__post_init__(**kwargs)


@dataclass
class Edge(OntologyElement):
    """
    A relationship between two nodes. We assume owlstar or similar for existential restrictions
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.Edge
    class_class_curie: ClassVar[str] = "om:Edge"
    class_name: ClassVar[str] = "edge"
    class_model_uri: ClassVar[URIRef] = OM.Edge

    subject: Optional[Union[str, NodeId]] = None
    predicate: Optional[Union[str, NodeId]] = None
    object: Optional[Union[str, NodeId]] = None
    subject_representation: Optional[str] = None
    predicate_representation: Optional[str] = None
    object_representation: Optional[str] = None
    annotation_set: Optional[Union[dict, Annotation]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.subject is not None and not isinstance(self.subject, NodeId):
            self.subject = NodeId(self.subject)

        if self.predicate is not None and not isinstance(self.predicate, NodeId):
            self.predicate = NodeId(self.predicate)

        if self.object is not None and not isinstance(self.object, NodeId):
            self.object = NodeId(self.object)

        if self.subject_representation is not None and not isinstance(self.subject_representation, str):
            self.subject_representation = str(self.subject_representation)

        if self.predicate_representation is not None and not isinstance(self.predicate_representation, str):
            self.predicate_representation = str(self.predicate_representation)

        if self.object_representation is not None and not isinstance(self.object_representation, str):
            self.object_representation = str(self.object_representation)

        if self.annotation_set is not None and not isinstance(self.annotation_set, Annotation):
            self.annotation_set = Annotation(**as_dict(self.annotation_set))

        super().__post_init__(**kwargs)


class LogicalDefinition(OntologyElement):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.LogicalDefinition
    class_class_curie: ClassVar[str] = "om:LogicalDefinition"
    class_name: ClassVar[str] = "logical definition"
    class_model_uri: ClassVar[URIRef] = OM.LogicalDefinition


class OntologySubset(OntologyElement):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.OntologySubset
    class_class_curie: ClassVar[str] = "om:OntologySubset"
    class_name: ClassVar[str] = "ontology subset"
    class_model_uri: ClassVar[URIRef] = OM.OntologySubset


# Enumerations
class OwlTypeEnum(EnumDefinitionImpl):

    CLASS = PermissibleValue(text="CLASS",
                                 meaning=OWL.Class)
    OBJECT_PROPERTY = PermissibleValue(text="OBJECT_PROPERTY",
                                                     meaning=OWL.ObjectProperty)
    NAMED_INDIVIDUAL = PermissibleValue(text="NAMED_INDIVIDUAL",
                                                       meaning=OWL.NamedIndividual)

    _defn = EnumDefinition(
        name="OwlTypeEnum",
    )

class SynonymScopeEnum(EnumDefinitionImpl):

    related = PermissibleValue(text="related",
                                     meaning=OIO.hasNarrowSynonym)
    broad = PermissibleValue(text="broad",
                                 meaning=OIO.hasBroadSynonym)
    narrow = PermissibleValue(text="narrow",
                                   meaning=OIO.hasNarrowSynonym)
    exact = PermissibleValue(text="exact",
                                 meaning=OIO.hasExactSynonym)

    _defn = EnumDefinition(
        name="SynonymScopeEnum",
    )

# Slots
class slots:
    pass

slots.owl_type = Slot(uri=OM.owl_type, name="owl type", curie=OM.curie('owl_type'),
                   model_uri=OM.owl_type, domain=None, range=Optional[Union[str, "OwlTypeEnum"]])

slots.name = Slot(uri=OM.name, name="name", curie=OM.curie('name'),
                   model_uri=OM.name, domain=None, range=Optional[str])

slots.subject = Slot(uri=OM.subject, name="subject", curie=OM.curie('subject'),
                   model_uri=OM.subject, domain=None, range=Optional[Union[str, NodeId]])

slots.object = Slot(uri=OM.object, name="object", curie=OM.curie('object'),
                   model_uri=OM.object, domain=None, range=Optional[Union[str, NodeId]])

slots.predicate = Slot(uri=OM.predicate, name="predicate", curie=OM.curie('predicate'),
                   model_uri=OM.predicate, domain=None, range=Optional[Union[str, NodeId]])

slots.annotation_set = Slot(uri=OM.annotation_set, name="annotation set", curie=OM.curie('annotation_set'),
                   model_uri=OM.annotation_set, domain=None, range=Optional[Union[dict, Annotation]])

slots.property = Slot(uri=OM.property, name="property", curie=OM.curie('property'),
                   model_uri=OM.property, domain=None, range=Optional[Union[str, NodeId]])

slots.filler = Slot(uri=OM.filler, name="filler", curie=OM.curie('filler'),
                   model_uri=OM.filler, domain=None, range=Optional[str])

slots.property_type = Slot(uri=OM.property_type, name="property type", curie=OM.curie('property_type'),
                   model_uri=OM.property_type, domain=None, range=Optional[str])

slots.filler_type = Slot(uri=OM.filler_type, name="filler type", curie=OM.curie('filler_type'),
                   model_uri=OM.filler_type, domain=None, range=Optional[str])

slots.subject_representation = Slot(uri=OM.subject_representation, name="subject representation", curie=OM.curie('subject_representation'),
                   model_uri=OM.subject_representation, domain=None, range=Optional[str])

slots.predicate_representation = Slot(uri=OM.predicate_representation, name="predicate representation", curie=OM.curie('predicate_representation'),
                   model_uri=OM.predicate_representation, domain=None, range=Optional[str])

slots.object_representation = Slot(uri=OM.object_representation, name="object representation", curie=OM.curie('object_representation'),
                   model_uri=OM.object_representation, domain=None, range=Optional[str])

slots.property_value_set = Slot(uri=OM.property_value_set, name="property value set", curie=OM.curie('property_value_set'),
                   model_uri=OM.property_value_set, domain=None, range=Optional[Union[Union[dict, PropertyValue], List[Union[dict, PropertyValue]]]])



import rdflib
from rdflib.namespace import (
    RDFS,
    RDF,
    OWL,
)
from rdflib import BNode, URIRef, Literal
from model.kgcl import (
    NodeRename,
    NodeObsoletion,
    NodeUnobsoletion,
    NodeDeletion,
    NodeMove,
    NodeDeepening,
    NodeShallowing,
    EdgeCreation,
    EdgeDeletion,
    PredicateChange,
    NodeCreation,
    ClassCreation,
    NewSynonym,
    RemovedNodeFromSubset,
)
from model.ontology_model import Edge
from diff.owlstar_sublanguage import (
    get_bnodes_2_atomic_existentials,
    get_bnodes_2_triple_annotations,
)
import time


def id_generator():
    """
    Returns a new ID for KGCL change operations.
    """
    id = 0
    while True:
        yield id
        id += 1


# initialise ID generator
id_gen = id_generator()

# TODO:
# obsoletion
# unobsoletion


def get_type(rdf_entity):
    if isinstance(rdf_entity, URIRef):
        return "uri"
    elif isinstance(rdf_entity, Literal):
        return "literal"
    else:
        return "Error"


def detect_renamings(added, deleted):
    """Given a diff represented by 'added' and 'deleted' triples,
    return an encoding in terms of
    (i) NodeRename instances,
    (ii) triples involved in (i), and
    (iii) nondeterministic choices.
    """

    covered = rdflib.Graph()

    # get labeling information
    added_labels = {}
    deleted_labels = {}

    for s, p, o in added.triples((None, RDFS.label, None)):
        if s not in added_labels:
            added_labels[s] = set()
        added_labels[s].add(o)

    for s, p, o in deleted.triples((None, RDFS.label, None)):
        if s not in deleted_labels:
            deleted_labels[s] = set()
        deleted_labels[s].add(o)

    # collect renamings
    nonDeterministic = []  # list of non-deterministic choices
    kgcl = []
    for subject in deleted_labels:
        if subject in added_labels:

            moved_from = deleted_labels[subject]
            moved_to = added_labels[subject]

            if len(moved_to) > 1 or len(moved_from) > 1:
                nonDeterministic.append((set(moved_from), set(moved_to)))

            shared = min(len(moved_to), len(moved_from))
            for x in range(shared):
                id = "test_id_" + str(next(id_gen))

                new = moved_to.pop()
                old = moved_from.pop()

                old_label = str(old)
                new_label = str(new)
                # get language tags
                old_language = old.language
                new_language = new.language

                node = NodeRename(
                    id=id,
                    about_node=subject,
                    old_value=old_label,
                    new_value=new_label,
                    old_language=old_language,
                    new_language=new_language,
                )
                kgcl.append(node)

                covered.add((subject, RDFS.label, new))
                covered.add((subject, RDFS.label, old))

    return kgcl, covered, nonDeterministic


def detect_node_moves(added, deleted):
    """Given a diff represented by 'added' and 'deleted' triples,
    return an encoding in terms of
    (i) NodeMove instances,
    (ii) triples involved in (i), and
    (iii) nondeterministic choices.
    """
    covered = rdflib.Graph()

    # map from subject 2 property (in added)
    s2p_added = {}
    for s, p, o in added:
        if s not in s2p_added:
            s2p_added[s] = set()
        s2p_added[s].add(p)

    # map from subject 2 property (in deleted)
    s2p_deleted = {}
    for s, p, o in deleted:
        if s not in s2p_deleted:
            s2p_deleted[s] = set()
        s2p_deleted[s].add(p)

    # identify triples that only differ wrt their object
    s2p_shared = {}
    for s in s2p_added:
        if s in s2p_deleted:
            s2p_shared[s] = s2p_added[s] & s2p_deleted[s]

    kgcl = []  # list of detected KGCL node modes
    nonDeterministic = []  # list of non-deterministic choices
    for subject in s2p_shared:
        for predicate in s2p_shared[subject]:

            # triples with the same subject and predicate
            moved_to = set()
            moved_from = set()
            for s, p, o in added.triples((subject, predicate, None)):
                if isinstance(o, URIRef):  # node move operates only on nodes
                    moved_to.add((s, p, o))
            for s, p, o in deleted.triples((subject, predicate, None)):
                if isinstance(o, URIRef):
                    moved_from.add((s, p, o))

            if len(moved_to) > 1 or len(moved_from) > 1:
                nonDeterministic.append((set(moved_from), set(moved_to)))

            # TODO: impose an order to make this deterministic
            shared = min(len(moved_to), len(moved_from))
            for x in range(shared):
                id = "test_id_" + str(next(id_gen))

                new = moved_to.pop()
                old = moved_from.pop()

                covered.add(new)
                covered.add(old)

                old_subject = str(old[0])
                old_predicate = str(old[1])
                old_object = str(old[2])
                new_object = str(new[2])

                edge = Edge(
                    subject=old_subject, predicate=old_predicate, object=old_object
                )

                # record entity type for KGCL rendering purposes
                old_object_type = get_type(old[2])
                new_object_type = get_type(new[2])

                node = NodeMove(
                    id=id,
                    about_edge=edge,
                    old_value=old_object,
                    new_value=new_object,
                    old_object_type=old_object_type,
                    new_object_type=new_object_type,
                )

                kgcl.append(node)

    return kgcl, covered, nonDeterministic


def detect_predicate_changes(added, deleted):
    """Given a diff represented by 'added' and 'deleted' triples,
    return an encoding in terms of
    (i) PredicateChange instances,
    (ii) triples involved in (i), and
    (iii) nondeterministic choices.
    """
    covered = rdflib.Graph()

    # maps from subjects to lists of objects
    s_2_os_deleted = {}
    s_2_os_added = {}

    for s, p, o in added:
        if s not in s_2_os_added:
            s_2_os_added[s] = []
        s_2_os_added[s].append(o)

    for s, p, o in deleted:
        if s not in s_2_os_deleted:
            s_2_os_deleted[s] = []
        s_2_os_deleted[s].append(o)

    # encode renamings wrt KGCL model
    kgcl = []
    nonDeterministic = []  # list of non-deterministic choices
    for s in s_2_os_added:
        if s in s_2_os_deleted:

            # get the objects as sets
            added_objects = set(s_2_os_added[s])
            deleted_objects = set(s_2_os_deleted[s])

            # get intersection of relevant objects
            shared_objects = added_objects & deleted_objects

            for i in shared_objects:
                # get corresponding predicates
                changed_to = set()
                changed_from = set()
                for s, p, o in added.triples((s, None, i)):
                    changed_to.add((s, p, o))
                for s, p, o in deleted.triples((s, None, i)):
                    changed_from.add((s, p, o))

                # create KGCL edge object
                edge = Edge(subject=str(s), object=str(i))

                language = None
                datatype = None
                if isinstance(i, Literal):
                    language = i.language
                    datatype = i.datatype
                    if datatype is not None:
                        datatype = (
                            "<" + datatype + ">"
                        )  # expect data types without curies

                # if i.language is not None:
                #    language = str(i.language)
                # else:
                #    language = None
                # if i.datatype is not None:
                #    datatype = str(i.datatype)
                # else:
                #    datatype = None

                object_type = get_type(i)

                if len(changed_to) > 1 or len(changed_from) > 1:
                    nonDeterministic.append((set(changed_from), set(changed_to)))

                # match potential predicate changes
                m = min(len(changed_to), len(changed_from))
                for x in range(m):
                    id = "test_id_" + str(next(id_gen))
                    old = (changed_from.pop())[1]
                    new = (changed_to.pop())[1]

                    # KGCL change object
                    change = PredicateChange(
                        id=id,
                        about_edge=edge,
                        old_value=str(old),
                        new_value=str(new),
                        object_type=object_type,
                        language=language,
                        datatype=datatype,
                    )

                    kgcl.append(change)
                    covered.add((s, old, i))
                    covered.add((s, new, i))

    return kgcl, covered, nonDeterministic



import click
import grammar.parser
import apply.graph_transformer
import rdflib

import sys

sys.path.append("../")
import python.kgcl


class Config(object):
    def __init__(self):
        self.verbose = False


pass_config = click.make_pass_decorator(Config, ensure=True)


@click.command()
@click.argument("graph", type=click.Path(), required=True)
@click.argument("kgcl", type=click.File("r"), required=True)
@click.argument("output", type=click.Path(), required=True)
# @click.option("--verbose", "-v", is_flag=True, help="Print more output.")
@pass_config
def cli(config, graph, kgcl, output):

    # read kgcl commands from file
    kgcl_patch = kgcl.read()

    # parser kgcl commands
    parsed_patch = grammar.parser.parse(kgcl_patch)

    # apply kgcl commands as SPARQL UPDATE queries to graph
    g = rdflib.Graph()
    # TODO: change this to 'parse' in order to support all formats
    # g.load(graph, format="nt")
    g.parse(graph)  # , format="nt") #TODO: this doesn't always work
    apply.graph_transformer.apply_patch(parsed_patch, g)

    # save updated graph
    g.serialize(destination=output, format="nt")


if __name__ == "__main__":
    cli()



from model.kgcl import (
    NodeRename,
    NodeObsoletion,
    NodeUnobsoletion,
    NodeDeletion,
    NodeMove,
    NodeDeepening,
    NodeShallowing,
    EdgeCreation,
    EdgeDeletion,
    PredicateChange,
    NodeCreation,
    ClassCreation,
    NewSynonym,
    PlaceUnder,
    RemoveUnder,
    RemovedNodeFromSubset,
    ExistentialRestrictionCreation,
    ExistentialRestrictionDeletion,
)
from model.ontology_model import Edge
import re


def render_entity(entity, rdf_type):
    entity = repr(entity)[1:-1]
    if rdf_type == "uri":
        return "<" + entity + ">"
    elif rdf_type == "label":
        if "'" in entity:
            # TODO: replacing quotes with backticks
            # is only a temporary workaround
            entity = entity.replace("'", "`")
        return "'" + entity + "'"
    elif rdf_type == "literal":
        # TODO: test this
        if '"' not in entity:
            return '"' + entity + '"'
        elif "'''" not in entity and entity[-1] != "'":
            return "'''" + entity + "'''"
        elif '"""' not in entity and entity[-1] != '"':
            return '"""' + entity + '"""'
        else:
            print("Rendering error: " + entity)
            raise
    else:
        print("Rendering error: " + entity)
        raise
    # return "'" + entity.replace("\\'", "`") + "'"


def render(kgclInstance):

    if type(kgclInstance) is NodeRename:
        # TODO: subject could be 'None'?
        subject = render_entity(kgclInstance.about_node, "uri")
        old = render_entity(kgclInstance.old_value, "label")
        new = render_entity(kgclInstance.new_value, "label")

        new_language = kgclInstance.new_language
        old_language = kgclInstance.old_language

        if old_language is not None:
            old = old + "@" + old_language

        if new_language is not None:
            new = new + "@" + new_language

        return "rename " + subject + " from " + old + " to " + new

    if type(kgclInstance) is NodeObsoletion:
        subject = render_entity(kgclInstance.about_node, "uri")
        # TODO: type this correctly
        replacement = render_entity(kgclInstance.has_direct_replacement, "uri")
        if kgclInstance.has_direct_replacement is not None:
            return "obsolete " + subject + " with replacement " + replacement
        else:
            return "obsolete " + subject

    if type(kgclInstance) is NodeUnobsoletion:
        subject = render_entity(kgclInstance.about_node, "uri")
        return "unobsolete " + subject

    if type(kgclInstance) is NodeDeletion:
        subject = render_entity(kgclInstance.about_node, "uri")
        return "delete " + subject

    if type(kgclInstance) is NodeMove:
        subject = render_entity(kgclInstance.about_edge.subject, "uri")
        predicate = render_entity(kgclInstance.about_edge.predicate, "uri")
        new = render_entity(kgclInstance.new_value, kgclInstance.new_object_type)
        old = render_entity(kgclInstance.old_value, kgclInstance.old_object_type)
        return (
            "move "
            + subject
            + " "
            + predicate
            + " "
            + old
            + " "
            + "from"
            + " "
            + old
            + " "
            + "to"
            + " "
            + new
        )

    if type(kgclInstance) is EdgeCreation:
        if kgclInstance.annotation_set is None:
            return render_edge_creation(kgclInstance)
        else:
            return render_annotation_creation(kgclInstance)

    if type(kgclInstance) is EdgeDeletion:
        if kgclInstance.annotation_set is None:
            return render_edge_deletion(kgclInstance)
        else:
            return render_annotation_deletion(kgclInstance)

    if type(kgclInstance) is PredicateChange:
        subject = render_entity(kgclInstance.about_edge.subject, "uri")
        object = render_entity(kgclInstance.about_edge.object, kgclInstance.object_type)
        new = render_entity(kgclInstance.new_value, "uri")
        old = render_entity(kgclInstance.old_value, "uri")

        if kgclInstance.language is not None:
            object += "@" + kgclInstance.language

        if kgclInstance.datatype is not None:
            object += "^^" + kgclInstance.datatype

        return (
            "change relationship between "
            + subject
            + " and "
            + object
            + " from "
            + old
            + " to "
            + new
        )

    if type(kgclInstance) is NodeCreation:
        subject = render_entity(kgclInstance.about_node, "uri")
        label = render_entity(kgclInstance.name, "label")
        if kgclInstance.name is not None:
            return "create node " + subject + " " + label
        else:
            return "create " + subject

    if type(kgclInstance) is ClassCreation:
        subject = render_entity(kgclInstance.about_node, "uri")
        return "create " + subject

    if type(kgclInstance) is NewSynonym:
        subject = render_entity(kgclInstance.about_node, "uri")
        synonym = render_entity(kgclInstance.new_value, "label")
        qualifier = kgclInstance.qualifier
        language = kgclInstance.language

        if language is not None:
            synonym = synonym + "@" + language

        if qualifier is not None:
            return (
                "create " + qualifier + " synonym" + " " + synonym + " for " + subject
            )
        else:
            return "create synonym " + synonym + " for " + subject

    if type(kgclInstance) is ExistentialRestrictionCreation:
        subclass = render_entity(kgclInstance.subclass, "uri")
        property = render_entity(kgclInstance.property, "uri")
        filler = render_entity(kgclInstance.filler, "uri")
        return "add " + subclass + " SubClassOf " + property + " some " + filler

    if type(kgclInstance) is ExistentialRestrictionDeletion:
        subclass = render_entity(kgclInstance.subclass, "uri")
        property = render_entity(kgclInstance.property, "uri")
        filler = render_entity(kgclInstance.filler, "uri")
        return "delete " + subclass + " SubClassOf " + property + " some " + filler

    if type(kgclInstance) is PlaceUnder:
        subclass = render_entity(kgclInstance.subject, "uri")
        superclass = render_entity(kgclInstance.object, "uri")
        return "add " + subclass + " SubClassOf " + superclass

    if type(kgclInstance) is RemoveUnder:
        subclass = render_entity(kgclInstance.subject, "uri")
        superclass = render_entity(kgclInstance.object, "uri")
        return "delete " + subclass + " SubClassOf " + superclass


def render_annotation_creation(kgclInstance):
    subject = render_entity(kgclInstance.subject, "uri")
    predicate = render_entity(kgclInstance.predicate, "uri")
    object = render_entity(kgclInstance.object, kgclInstance.object_type)

    annotation = kgclInstance.annotation_set
    annotation_property = render_entity(annotation.property, "uri")
    annotation_filler = render_entity(annotation.filler, annotation.filler_type)

    language = kgclInstance.language
    datatype = kgclInstance.datatype

    if language is not None:
        object = object + "@" + language
    if datatype is not None:
        object = object + "^^" + datatype

    return (
        "create edge <<"
        + subject
        + " "
        + predicate
        + " "
        + object
        + ">>"
        + " "
        + annotation_property
        + " "
        + annotation_filler
    )


def render_annotation_deletion(kgclInstance):
    subject = render_entity(kgclInstance.subject, "uri")
    predicate = render_entity(kgclInstance.predicate, "uri")
    object = render_entity(kgclInstance.object, kgclInstance.object_type)

    annotation = kgclInstance.annotation_set
    annotation_property = render_entity(annotation.property, "uri")
    annotation_filler = render_entity(annotation.filler, annotation.filler_type)

    language = kgclInstance.language
    datatype = kgclInstance.datatype

    if language is not None:
        object = object + "@" + language
    if datatype is not None:
        object = object + "^^" + datatype

    return (
        "delete edge <<"
        + subject
        + " "
        + predicate
        + " "
        + object
        + ">>"
        + " "
        + annotation_property
        + " "
        + annotation_filler
    )


def render_edge_deletion(kgclInstance):
    subject = render_entity(kgclInstance.subject, "uri")
    predicate = render_entity(kgclInstance.predicate, "uri")
    object = render_entity(kgclInstance.object, kgclInstance.object_type)
    # object = render_entity(repr(kgclInstance.object)[1:-1])

    language = kgclInstance.language
    datatype = kgclInstance.datatype

    base = "delete edge " + subject + " " + predicate + " " + object

    if language is not None:
        return base + "@" + language
    elif datatype is not None:
        return base + "^^" + datatype
    else:
        return base


def render_edge_creation(kgclInstance):
    subject = render_entity(kgclInstance.subject, "uri")
    predicate = render_entity(kgclInstance.predicate, "uri")
    object = render_entity(kgclInstance.object, kgclInstance.object_type)

    language = kgclInstance.language
    datatype = kgclInstance.datatype

    base = "create edge " + subject + " " + predicate + " " + object

    if language is not None:
        return base + "@" + language
    elif datatype is not None:
        return base + "^^" + datatype
    else:
        return base



import rdflib
from .owlstar_sublanguage import (
    get_thin_triples,
    get_atomic_existentials,
    get_triple_annotations,
    get_atomic_subsumptions,
)


def get_added_thin_triples(g1, g2):
    """
    Given two graphs g1 and g2,
    returns triples (without blank nodes) that are in g2 but not g1,
    i.e. 'added' triples.
    """
    thin1 = get_thin_triples(g1)
    thin2 = get_thin_triples(g2)

    return thin2 - thin1


def get_deleted_thin_triples(g1, g2):
    """
    Given two graphs g1 and g2,
    returns triples (without blank nodes) that are in g1 but not g2,
    i.e. 'deleted' triples.
    """
    thin1 = get_thin_triples(g1)
    thin2 = get_thin_triples(g2)

    return thin1 - thin2


def get_added_subsumptions(g1, g2):
    """
    Given two graphs g1 and g2,
    returns atomic subsumption axioms that are in g2 but not g1,
    i.e. 'added' subsumptions.
    """
    sub1 = get_atomic_subsumptions(g1)
    sub2 = get_atomic_subsumptions(g2)

    return sub2 - sub1


def get_deleted_subsumptions(g1, g2):
    """
    Given two graphs g1 and g2,
    returns atomic subsumption axioms that are in g1 but not g2,
    i.e. 'deleted' subsumptions.
    """
    sub1 = get_atomic_subsumptions(g1)
    sub2 = get_atomic_subsumptions(g2)

    return sub1 - sub2


# NB: this returns a set of 'ExistentialRestrictions'
def get_added_existentials(g1, g2):
    """
    Given two graphs g1 and g2,
    returns ExistentialRestrictions that are in g2 but not g1,
    i.e. 'added' ExistentialRestrictions.
    """
    ex1 = set(get_atomic_existentials(g1))
    ex2 = set(get_atomic_existentials(g2))

    return ex2 - ex1


def get_deleted_existentials(g1, g2):
    """
    Given two graphs g1 and g2,
    returns ExistentialRestrictions that are in g1 but not g2,
    i.e. 'deleted' ExistentialRestrictions.
    """
    ex1 = set(get_atomic_existentials(g1))
    ex2 = set(get_atomic_existentials(g2))

    return ex1 - ex2


# NB this returns a set of 'TripleAnnotation's
def get_added_triple_annotations(g1, g2):
    """
    Given two graphs g1 and g2,
    returns ExistentialRestrictions that are in g2 but not g1,
    i.e. 'added' ExistentialRestrictions.
    """
    annotations1 = set(get_triple_annotations(g1))
    annotations2 = set(get_triple_annotations(g2))

    return annotations2 - annotations1


def get_deleted_triple_annotations(g1, g2):
    """
    Given two graphs g1 and g2,
    returns TripleAnnotations that are in g1 but not g2,
    i.e. 'deleted' ExistentialRestrictions.
    """
    annotations1 = set(get_triple_annotations(g1))
    annotations2 = set(get_triple_annotations(g2))

    return annotations1 - annotations2


if __name__ == "__main__":
    g1 = rdflib.Graph()
    g2 = rdflib.Graph()

    g1.load("n3/obi/obi_1.nt", format="nt")
    g2.load("n3/obi/obi_2.nt", format="nt")

    deleted = get_deleted_triples(g1, g2)
